Index: include/openbios/drivers.h
===================================================================
--- include/openbios/drivers.h	(revision 54)
+++ include/openbios/drivers.h	(working copy)
@@ -15,7 +15,7 @@
 int ob_pci_init(void);
 #endif
 #ifdef CONFIG_DRIVER_SBUS
-int ob_sbus_init(void);
+int ob_sbus_init(unsigned long base);
 #endif
 #ifdef CONFIG_DRIVER_IDE
 int ob_ide_init(void);
Index: arch/sparc32/console.c
===================================================================
--- arch/sparc32/console.c	(revision 54)
+++ arch/sparc32/console.c	(working copy)
@@ -41,9 +41,12 @@
 #define	BRENAB 	1	/* Baud rate generator enable */
 #define	BRSRC	2	/* Baud rate generator source */
 
+/* Read Register 0 */
+#define	Rx_CH_AV	0x1	/* Rx Character Available */
+
 static int uart_charav(int port)
 {
-	return ((inb(CTRL(port)) & 1) != 0);
+	return ((inb(CTRL(port)) & Rx_CH_AV) != 0);
 }
 
 static char uart_getchar(int port)
Index: arch/sparc32/openbios.c
===================================================================
--- arch/sparc32/openbios.c	(revision 54)
+++ arch/sparc32/openbios.c	(working copy)
@@ -21,8 +21,8 @@
 void tcx_init(unsigned long base);
 
 #define IOMMU_BASE    0x10000000 /* First page of sun4m IOMMU */
+#define SBUS_BASE     0x10001000
 #define SLAVIO_BASE   0x71000000
-#define MACIO_BASE    0x70000000
 #define TCX_BASE      0x50000000
 
 static unsigned char intdict[256 * 1024];
@@ -49,15 +49,12 @@
 	modules_init();
 #ifdef CONFIG_DRIVER_SBUS
         ob_init_mmu(IOMMU_BASE);
-	ob_sbus_init();
+	ob_sbus_init(SBUS_BASE);
 
 #ifdef CONFIG_DEBUG_CONSOLE_VIDEO
 	init_video();
 #endif
 #endif
-#ifdef CONFIG_DRIVER_ESP
-	ob_esp_init(MACIO_BASE);
-#endif
 #ifdef CONFIG_DRIVER_OBIO
 	ob_obio_init(SLAVIO_BASE);
         nvram_init();
Index: arch/sparc32/tree.fs
===================================================================
--- arch/sparc32/tree.fs	(revision 54)
+++ arch/sparc32/tree.fs	(working copy)
@@ -58,14 +58,12 @@
   h# 01443fd0 encode-int " clock-frequency" property
   h# 1c encode-int " slot-address-bits" property
   h# 3f encode-int " burst-sizes" property
-  h# ffee7000 encode-int " address" property
   h# 0 encode-int h# 0 encode-int encode+ h# 0 encode-int encode+ h# 30000000 encode-int encode+ h# 10000000 encode-int encode+
    h# 1 encode-int encode+ h# 0 encode-int encode+ h# 0 encode-int encode+ h# 40000000 encode-int encode+ h# 10000000 encode-int encode+
    h# 2 encode-int encode+ h# 0 encode-int encode+ h# 0 encode-int encode+ h# 50000000 encode-int encode+ h# 10000000 encode-int encode+
    h# 3 encode-int encode+ h# 0 encode-int encode+ h# 0 encode-int encode+ h# 60000000 encode-int encode+ h# 10000000 encode-int encode+
    h# 4 encode-int encode+ h# 0 encode-int encode+ h# 0 encode-int encode+ h# 70000000 encode-int encode+ h# 10000000 encode-int encode+
    " ranges" property
-  h# 0 encode-int h# 10001000 encode-int encode+ h# 00000028 encode-int encode+ " reg" property
   external
   : open ( cr ." opening SBus" cr) true ;
   : close ;
Index: arch/sparc32/romvec.c
===================================================================
--- arch/sparc32/romvec.c	(revision 54)
+++ arch/sparc32/romvec.c	(working copy)
@@ -367,18 +367,21 @@
     return ret;
 }
 
-static void obp_dumb_memfree(char *va, unsigned sz)
+static void obp_dumb_memfree(__attribute__((unused))char *va,
+                             __attribute__((unused))unsigned sz)
 {
     DPRINTF("obp_dumb_memfree 0x%x(%d)\n", va, sz);
 }
 
 static char * obp_dumb_memalloc(char *va, unsigned size)
 {
-    static char *pa = 0x4000000;
+    char *va2;
 
     DPRINTF("obp_dumb_memalloc req 0x%x of %d at 0x%x\n", va, size, pa);
-    obp_dumb_mmap(va, 1, pa, size);
-    pa += size;
+    
+    va2 = malloc(size);
+    obp_dumb_mmap(va, 1, va2pa((unsigned long)va2), size);
+
     return va;
 }
 
Index: drivers/sbus.c
===================================================================
--- drivers/sbus.c	(revision 54)
+++ drivers/sbus.c	(working copy)
@@ -18,7 +18,77 @@
 
 #include "openbios/drivers.h"
 
+#define SBUS_REGS        0x28
+#define SBUS_SLOTS       5
+
 static void
+ob_sbus_node_init(unsigned long base)
+{
+    void *regs;
+
+    push_str("/iommu/sbus");
+    fword("find-device");
+
+    PUSH(0);
+    fword("encode-int");
+    PUSH(base);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(SBUS_REGS);
+    fword("encode-int");
+    fword("encode+");
+    push_str("reg");
+    fword("property");
+
+    regs = map_io(base, SBUS_REGS);
+    PUSH((unsigned long)regs);
+    fword("encode-int");
+    push_str("address");
+    fword("property");
+}
+
+static void
+ob_macio_init(unsigned int slot, unsigned long base)
+{
+    // All devices were integrated to NCR89C100, see
+    // http://www.ibiblio.org/pub/historic-linux/early-ports/Sparc/NCR/NCR89C100.txt
+
+    // NCR 53c9x, aka ESP. See
+    // http://www.ibiblio.org/pub/historic-linux/early-ports/Sparc/NCR/NCR53C9X.txt
+#ifdef CONFIG_DRIVER_ESP
+    ob_esp_init(base);
+#endif
+
+    // NCR 92C990, Am7990, Lance. See http://www.amd.com
+    //ob_le_init(base);
+
+    // Parallel port
+    //ob_bpp_init(base);
+
+    // Power management
+    //ob_power_init(base);
+}
+
+static void
+sbus_probe_slot(unsigned int slot, unsigned long base)
+{
+    // OpenBIOS and Qemu don't know how to do Sbus probing
+    switch(slot) {
+    case 2: // SUNW,tcx
+        //ob_tcx_init(slot, base);
+        break;
+    case 3: // SUNW,CS4231
+        //ob_cs4231_init(slot, base);
+        break;
+    case 4: // MACIO: le, esp, bpp, power-management
+        ob_macio_init(slot, base);
+        break;
+    default:
+        break;
+    }
+}
+
+static void
 ob_sbus_open(int *idx)
 {
 	int ret=1;
@@ -43,10 +113,23 @@
 	{ "close",		ob_sbus_close		},
 };
 
+static const unsigned long sbus_offset[SBUS_SLOTS] = {
+    0x30000000,
+    0x40000000,
+    0x50000000,
+    0x60000000,
+    0x70000000,
+};
 
-int ob_sbus_init(void)
+int ob_sbus_init(unsigned long base)
 {
-        //printk("Initializing SBus devices...\n");
-	
-	return 0;
+    unsigned int slot;
+
+    ob_sbus_node_init(base);
+
+    for (slot = 0; slot < SBUS_SLOTS; slot++) {
+        sbus_probe_slot(slot, sbus_offset[slot]);
+    }
+
+    return 0;
 }
Index: drivers/iommu.c
===================================================================
--- drivers/iommu.c	(revision 54)
+++ drivers/iommu.c	(working copy)
@@ -18,6 +18,9 @@
 #include "pgtsrmmu.h"
 #include "iommu.h"
 
+#define IOMMU_REGS  0x300
+#define NCTX_SWIFT  0x100
+
 #define IOPERM        (IOPTE_CACHE | IOPTE_WRITE | IOPTE_VALID)
 #define MKIOPTE(phys) (((((phys)>>4) & IOPTE_PAGE) | IOPERM) & ~IOPTE_WAZ)
 #define LOWMEMSZ 32 * 1024 * 1024
@@ -57,14 +60,14 @@
 };
 
 struct iommu ciommu;
+static struct iommu_regs *regs;
 
-#define NCTX_SWIFT  0x100
-
 static void iommu_init(struct iommu *t, unsigned long base);
 
 static void
-iommu_invalidate(struct iommu_regs *regs) {
-    regs->tlbflush = 0;
+iommu_invalidate(struct iommu_regs *iregs)
+{
+    iregs->tlbflush = 0;
 }
 
 /*
@@ -233,7 +236,7 @@
     PUSH(base);
     fword("encode-int");
     fword("encode+");
-    PUSH(0x300);
+    PUSH(IOMMU_REGS);
     fword("encode-int");
     fword("encode+");
     push_str("reg");
@@ -259,8 +262,16 @@
     fword("encode+");
     push_str("available");
     fword("property");
+
+    push_str("/iommu");
+    fword("find-device");
+    PUSH((unsigned long)regs);
+    fword("encode-int");
+    push_str("address");
+    fword("property");
 }
 
+
 /*
  * Switch page tables.
  */
@@ -381,11 +392,11 @@
 {
     unsigned int *ptab;
     int ptsize;
-    struct iommu_regs *regs =NULL;
     unsigned int impl, vers;
     unsigned int tmp;
 
-    if ((regs = map_io(base, sizeof(struct iommu_regs))) == 0) {
+    regs = map_io(base, IOMMU_REGS);
+    if (regs == 0) {
         DPRINTF("Cannot map IOMMU\n");
         for (;;) { }
     }
Index: drivers/obio.c
===================================================================
--- drivers/obio.c	(revision 54)
+++ drivers/obio.c	(working copy)
@@ -537,6 +537,9 @@
 ob_obio_init(unsigned long slavio_base)
 {
 
+    // All devices were integrated to NCR89C105, see
+    // http://www.ibiblio.org/pub/historic-linux/early-ports/Sparc/NCR/NCR89C105.txt
+
     //printk("Initializing OBIO devices...\n");
 #if 0 // XXX
     REGISTER_NAMED_NODE(ob_obio, "/obio");
@@ -544,13 +547,16 @@
     ob_set_obio_ranges(slavio_base);
 #endif
 
+    // Zilog Z8530 serial ports, see http://www.zilog.com
     // Must be before zs@0,0 or Linux won't boot
     ob_zs_init(slavio_base, SLAVIO_ZS1, ZS_INTR, 0, 0);
 
     ob_zs_init(slavio_base, SLAVIO_ZS, ZS_INTR, 1, 1);
 
+    // M48T08 NVRAM, see http://www.st.com
     ob_nvram_init(slavio_base, SLAVIO_NVRAM);
 
+    // 82078 FDC
     ob_fd_init(slavio_base, SLAVIO_FD, FD_INTR);
 
     ob_sconfig_init(slavio_base, SLAVIO_SCONFIG);
